#!/usr/bin/env python3

# Class used to help send data to statsd
# Automatically includes things like dimensions to the output to make sending stats easier

import sys
import os
import socket
import json
import statsd
import argparse

# if script's parent directory isn't in sys.path, then add it
if ('/'.join(sys.path[0].split("/")[:-1]) not in sys.path):
    sys.path.append('/'.join(sys.path[0].split("/")[:-1]))

from apsdk import apenv


class StatsdHelper(object):
    def __init__(self,
                 statsd_port: int,
                 statsd_host: str="localhost",
                 test_status: bool=False,
                 test_data: dict={}):

        self.test_status = test_status
        self.statsd_host = statsd_host
        self.statsd_port = statsd_port

        self.hostname = socket.gethostname()
        self.metric = statsd.StatsClient(statsd_host, statsd_port)

        # If this is a normal run we get our data from localwatchdogstatus.csv
        # and from apenv functions
        if not self.test_status:
            self.input_csv_file = apenv.get_data_root() + "/watchdogoutput/localwatchdogstatus.csv"

            # Tenant = AP Cluster
            # Role = AP Environment
            # RoleInstance = AP Machine Function
            self.tenant = apenv.get_cluster_name()
            self.role = apenv.get_environment_name()
            self.role_instance = apenv.get_machine_function()

        # If this is a unit test we get our data from test data passed in
        else:
            self.input_csv_file = test_data["input_csv_file"]
            self.tenant = test_data["Tenant"]
            self.role = test_data["Role"]
            self.role_instance = test_data["RoleInstance"]

        self.osversion = ""
        self.shared_services_version = ""

        # Call get_version_info method to get OS version and shared services version
        self.get_version_info()

    # Get version information for the OS and shared services
    def get_version_info(self) -> bool:
        # Verify that watchdog has already created this file
        if not os.path.isfile(self.input_csv_file):
            return False

        try:
            with open(self.input_csv_file, 'r') as infile:
                for line in infile:
                    if "OSImageVersion" in line:
                        self.osversion = line.split(",")[3].strip()
                    elif "aplinux-shared-services" in line:
                        self.shared_services_version = line.split(",")[3].strip()
        except Exception as e:
            print("statsdhelper.get_version_info: Exception: " + str(e))
            return False

        # If the values are not present yet in the file we can error out and try again later
        # once the watchapversion.py watchdog runs and gets this info
        if not (self.osversion and self.shared_services_version):
            return False

        return True

    # Send gauge statistic to statsd with required Dimensions
    # Minimal input required to send a metric
    def send_gauge(self, namespace: str, metric_name: str, metric_value: int, custom_dims: dict={}) -> bool:
        # print("\nSending statsd values:")
        # print("statsdhelper.send_gauge Namespace: " + namespace)
        # print("statsdhelper.send_gauge Metric: " + metric_name)
        # print("statsdhelper.send_gauge Tenant: " + self.tenant)
        # print("statsdhelper.send_gauge Role: " + self.role)
        # print("statsdhelper.send_gauge RoleInstance: " + self.role_instance)
        # print("statsdhelper.send_gauge MachineName: " + self.hostname)
        # print("statsdhelper.send_gauge OSVer: " + self.osversion)
        # print("statsdhelper.send_gauge SvcVer: " + self.shared_services_version)
        # print("statsdhelper.send_gauge Value: " + str(metric_value))

        # Make sure all required dimensions are present and accounted for
        if not all((namespace, metric_name, metric_value, self.tenant, self.role, self.role_instance,
                    self.hostname, self.osversion, self.shared_services_version)):
            print("statsdhelper.send_gauge: Not all of the required dimensions were defined properly")
            return False

        dims_dict = {"Tenant": self.tenant,
                     "Role": self.role,
                     "RoleInstance": self.role_instance,
                     "MachineName": self.hostname,
                     "OSVer": self.osversion,
                     "SvcVer": self.shared_services_version}

        # Add custom dimensions
        if type(custom_dims) is dict:
            dims_dict.update(custom_dims)

        metric_dict = {"Namespace": namespace,
                       "Metric": metric_name,
                       "Dims": dims_dict}

        # Show JSON output sent to statsd
        # print("\nJSON Formatted output for statsd:")
        # print(json.dumps(metric_dict, indent=4, sort_keys=True))

        try:
            metric_json = json.dumps(metric_dict)
        except Exception as e:
            print("statsdhelper.send_gauge: json exception: " + str(e))
            return False

        try:
            self.metric.gauge(metric_json, metric_value)
        except Exception as e:
            print("statsdhelper.send_gauge: statsd exception: " + str(e))
            return False

        return True

# Example run command:
# /ap/app/aptools/statsdhelper.py -p 8125 -n MyNamespace -m ServiceFailures -a 1
# /ap/app/aptools/statsdhelper.py -p 8125 -n MyNamespace -m ServiceFailures -a 1 -c "Cause:DBTimeout,User:dbadmin"

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Used to send metrics to Geneva Hot Path (MDM) "
                                     "with additional, useful Autopilot environment dimensions.")
    parser.add_argument("-s", "--statsdhost", metavar="STATSDHOST", default="localhost",
                        help="statsd host to send metrics through, defaults to localhost")
    parser.add_argument("-p", "--port", metavar="PORT", help="statsd port to send metrics through")
    parser.add_argument("-n", "--namespace", metavar="NAMESPACE", help="MDM namespace")
    parser.add_argument("-m", "--metricname", metavar="METRICNAME", help="metric name")
    parser.add_argument("-a", "--metricvalue", metavar="METRICVALUE", help="metric value (integer)")
    parser.add_argument("-c", "--customdims", metavar="DIMENSIONS",
                        help="comma-delimited list of additional dimensions to send with the "
                             "metric, with the dimension name and value separated by a colon "
                             "(optional)")

    args = parser.parse_args()

    if not (args.statsdhost and args.port and args.namespace and args.metricname and args.metricvalue):
        print("ERROR: Missing required parameters", file=sys.stderr)
        parser.print_help()
        sys.exit(1)

    # parse statsd port number
    try:
        port = int(args.port)
    except:
        print("ERROR: statsd port number must be an integer", file=sys.stderr)
        parser.print_help()
        sys.exit(1)

    if port < 1 or port > 65535:
        print("ERROR: statsd port number must be between 1 and 65535", file=sys.stderr)
        parser.print_help()
        sys.exit(1)

    # parse the metric value
    try:
        metricvalue = int(args.metricvalue)
    except:
        print("ERROR: metric value must be an integer", file=sys.stderr)
        parser.print_help()
        sys.exit(1)

    # parse custom dimensions
    customdims = {}
    if args.customdims:
        for dim in args.customdims.split(","):
            d = dim.split(":")
            if len(d) != 2:
                print("ERROR: syntax error in custom dimension '{0}'".format(dim), file=sys.stderr)
                parser.print_help()
                sys.exit(1)

            k = d[0].strip()
            v = d[1].strip()
            if not (k and v):
                print("ERROR: syntax error in custom dimension '{0}'".format(dim), file=sys.stderr)
                parser.print_help()
                sys.exit(1)

            customdims[k] = v

    metrics = StatsdHelper(statsd_host=args.statsdhost, statsd_port=port)
    if not metrics.send_gauge(args.namespace, args.metricname, metricvalue, customdims):
        print("ERROR: failed to send metric to mdm", file=sys.stderr)
        sys.exit(1)
