#!/bin/bash
set -exuo pipefail

DNS_CONFIG="/etc/dnsmasq.d/apdocker-dnsmasq.conf"
DNS_SUFFIX="apdocker.ap.gbl"
TM_URL="aplinuxservices.trafficmanager.net"

SYSTEMD_UNIT_ROOT="/lib/systemd/system"

MDM_PORT="9125"
MDM_NAMESPACE="APSharedServices"
MDM_METRIC_SUCCESS="TmApdockerS"
MDM_METRIC_FAIL="TmApdockerF"

# Sends a metric to Geneva Hot Path (MDM). The first parameter is the statsd port,
# the second parameter is the MDM namespace, the third the metric name, the fourth
# is the metric value, and the fifth is an optional comma-delimited list of
# dimensions (key-value pairs) to send with the metric. The dimension name and
# value are separated by a colon (:).
function send_gauge {
    local cmd="/ap/app/aptools/statsdhelper.py -p ${1} -n ${2} -m ${3} -a ${4} -c ${5}"

    eval ${cmd}
    local ret_code=$?

    if [ ${ret_code} -ne 0 ]; then
        echo "Traffic Manager Monitor (${TM_URL}): Error sending metric ${2} to MDM namespace ${1} using statsd port ${MDM_PORT}" >&2
    fi
}

# Verify that service exists
function check_service_exists {
    if [ -f ${2}"/"${1}".service" ]; then
        echo "1"
    else
        echo "0"
    fi
}

# Get the ActiveState status of the service being acted upon to make sure it's started
function get_service_activestate {
    echo "Traffic Manager Monitor (${TM_URL}): Getting ActiveState for ${1}.service" >&2
    local result=$(systemctl show ${1}.service | egrep 'ActiveState')
    IFS='=' read -r -a activestate_array <<< "${result}"
    local activestate="${activestate_array[1]}"
    echo "Traffic Manager Monitor (${TM_URL}): ActiveState for ${1}.service is ${activestate}" >&2

    echo ${activestate}
}

# Get the SubState status of the service being acted upon to make sure it's started
function get_service_substate {
    echo "Traffic Manager Monitor (${TM_URL}): Getting SubState for ${1}.service" >&2
    local result=$(systemctl show ${1} | egrep 'SubState')
    IFS='=' read -r -a substate_array <<< "${result}"
    local substate="${substate_array[1]}"
    echo "Traffic Manager Monitor (${TM_URL}): SubState for ${1}.service is ${substate}" >&2

    echo ${substate}
}

# Get the UnitFileState of the service being acted upon to make sure it's enabled
function get_service_unitfilestate {
    echo "Traffic Manager Monitor (${TM_URL}): Getting UnitFileState for ${1}.service" >&2
    local result=$(systemctl show ${1} | egrep 'UnitFileState')
    IFS='=' read -r -a unitfilestate_array <<< "${result}"
    local unitfilestate="${unitfilestate_array[1]}"
    echo "Traffic Manager Monitor (${TM_URL}): UnitFileState for ${1}.service is ${unitfilestate}" >&2

    echo ${unitfilestate}
}

# Start the service, restarting it if its already running
function start_service {
    echo "Traffic Manager Monitor (${TM_URL}): Attempting to start ${1}.service" >&2

    local service_exists=$(check_service_exists "${1}" "${SYSTEMD_UNIT_ROOT}")
    if [ ${service_exists} -eq 1 ]; then
        echo ": Service ${1}.service exists on system" >&2
        local attempt=1
        while true;
        do
            echo "Traffic Manager Monitor (${TM_URL}): Starting ${1}.service attempt ${attempt}" >&2
            # Make sure we really stopped before starting, otherwise will need to wait till
            # the next trigger is issued. Ignore error since the service may have already
            # stopped.
            systemctl stop ${1}.service || true
            local cmd="systemctl start ${1}.service"

            eval ${cmd}
            local ret_code=$?

            # After running stop the service should be in the following state:

            # ActiveState=active or activating
            # SubState=running or auto-restart or exited

            # We verify to make sure ActiveState and SubState match the above states
            # Anything else is a failure and we exit install

            local activestate=$(get_service_activestate "${1}")
            local substate=$(get_service_substate "${1}")

            echo "Traffic Manager Monitor (${TM_URL}): Service ActiveState = ${activestate}" >&2
            echo "Traffic Manager Monitor (${TM_URL}): Service SubState = ${substate}" >&2

            # Determine if service is active
            if [ ${activestate} != "active" ] && [ ${activestate} != "activating" ]; then
                activestate_status="inactive"
            elif [ ${activestate} == "active" ] || [ ${activestate} == "activating" ]; then
                activestate_status="active"
            else
                activestate_status="unknown"
            fi

            # Determine if service has started
            if [ ${substate} != "running" ] && [ ${substate} != "auto-restart" ] && [ ${substate} != "exited" ]; then
                substate_status="stopped"
            elif [ ${substate} == "running" ] || [ ${substate} == "auto-restart" ] || [ ${substate} == "exited" ]; then
                substate_status="running"
            else
                substate_status="unknown"
            fi

            # Combine comparisons to make sure service is started; if not, attempt to restart it again
            if [ ${ret_code} -ne 0 ] && [ ${activestate_status} == "inactive" ] && [ $substate_status == "stopped" ]; then
                attempt=$(expr ${attempt} + 1)
            elif [ ${activestate_status} == "active" ] && [ ${substate_status} == "running" ]; then
                echo "Traffic Manager Monitor (${TM_URL}): Successfully Started ${1}.service" >&2
                break
            else
                echo "Traffic Manager Monitor (${TM_URL}): Unusual state reached while starting ${1}.service" >&2
                send_gauge ${MDM_PORT} ${MDM_NAMESPACE} ${MDM_METRIC_FAIL} 1 "Reason:BadState"
                exit 1
            fi

            if [ ${attempt} -le 4 ]; then
                sleep 10
            else
                echo "Traffic Manager Monitor (${TM_URL}): Failed to start ${1}.service after 4 tries." >&2
                send_gauge ${MDM_PORT} ${MDM_NAMESPACE} ${MDM_METRIC_FAIL} 1 "Reason:NoStart"
                exit 1
            fi
        done
    else
        echo "Traffic Manager Monitor (${TM_URL}): Service ${1}.service does not exist on system: Failure!" >&2
        send_gauge ${MDM_PORT} ${MDM_NAMESPACE} ${MDM_METRIC_FAIL} 1 "Reason:NoService"
        exit 1
    fi
}

PRIOR_IP=$(awk -F '=' '/\/'${DNS_SUFFIX}'\// {print $2}' ${DNS_CONFIG} | cut -d'/' -f3)
TM_IP=$(getent hosts ${TM_URL} | awk '{print $1}')

if [ "${PRIOR_IP}" == "${TM_IP}" ]; then
    echo "Traffic Manager Monitor (${TM_URL}): No change detected" >&2
    send_gauge ${MDM_PORT} ${MDM_NAMESPACE} ${MDM_METRIC_SUCCESS} 1 "Reason:NoChange"
    exit 0
fi

echo "Traffic Manager Monitor (${TM_URL}): Change detected in IP from ${PRIOR_IP} to ${TM_IP}" >&2

# update dnsmasq config file
sed -i "s/^address\=\/${DNS_SUFFIX}\/.*/address=\/${DNS_SUFFIX}\/${TM_IP}/" ${DNS_CONFIG}

# restart dnsmasq
start_service dnsmasq

# resolve ip
#   awk command finds the line starting with 'Name:' and prints the second
#   colon-delimited field of the next line
RESOLVED_IP=$(nslookup "test.${DNS_SUFFIX}" | awk '{if (c&&!--c){print $2}};/^Name:/{c=1}')
if [ "${RESOLVED_IP}" != "${TM_IP}" ]; then
    echo "Traffic Manager Monitor (${TM_URL}): Failed resolve to new traffic manager ip. Got ${RESOLVED_IP} but expected ${TM_IP}" >&2
    send_gauge ${MDM_PORT} ${MDM_NAMESPACE} ${MDM_METRIC_FAIL} 1 "Reason:ResolutionFail"
    exit 1
fi

send_gauge ${MDM_PORT} ${MDM_NAMESPACE} ${MDM_METRIC_SUCCESS} 1 "Reason:NewAddress"

