#!/usr/bin/env python3

# Watchdog used to check service statuses and report on them

import sys
import os
import json
import subprocess
import io
import platform
import socket

# if script's parent directory isn't in sys.path, then add it
if ('/'.join(sys.path[0].split("/")[:-1]) not in sys.path):
    sys.path.append('/'.join(sys.path[0].split("/")[:-1]))

from aptools import statsdhelper


class WatchAPServiceStatus(object):
    def __init__(self, configpath: str, watchdog_csv_dir: str, watchdog_tmp_file: str, watchdog_csv_file: str,
                 test_status=False):

        # If we are in unit test mode we deal with certain situations
        # which we can't replicate with generic data
        self.test_status = test_status

        self.configpath = configpath

        # Suffix of files to read from configuration directory
        self.suffix = ".conf"

        # Store JSON service configs in memory
        self.json_configs = {}  # All the JSON configs loaded by the watchdog

        # Store reported status per service in dict
        self.reported_status = {}

        # machinelocalwatchdog folder and files to use for DM report
        self.watchdog_csv_dir = watchdog_csv_dir
        self.watchdog_tmp_file = watchdog_tmp_file
        self.watchdog_csv_file = watchdog_csv_file

        # Get hostname
        self.hostname = socket.gethostname()

        # MDM Namespace to report stats to
        self.namespace = "APSharedServices"

        # Only load statsdhelper for MDM reporting if we're NOT doing unit tests
        if not self.test_status:
            self.metrics = statsdhelper.StatsdHelper(statsd_port=9125)

    # Read in .conf files from conf.d folder of watchapservicestatus
    def read_conf_files(self) -> bool:
        for file in os.listdir(self.configpath):
            if file.endswith(self.suffix):
                try:
                    with open(os.path.join(self.configpath, file)) as data_file:
                        self.json_configs[os.path.join(self.configpath, file)] = json.load(data_file)

                except Exception as e:
                    print("watchapservicestatus.read_conf_files: Exception: " + str(e))
                    return False

        if len(self.json_configs) < 1:
            return False

        return True

    # Print out the JSON configuration
    def print_conf_files(self) -> bool:
        try:
            print("\nService Configurations:")
            print(json.dumps(self.json_configs, indent=4, sort_keys=True))
        except Exception as e:
            print("watchapservicestatus.print_dm_report: Exception: " + str(e))
            return False

        return True

    # Make sure that certain JSON values are present and that they match certain expected values
    def verify_conf_files(self) -> bool:
        for key, value in self.json_configs.items():
            try:
                service_config_source = self.json_configs[key]["config"]["statusSource"]
            except KeyError:
                print("watchapservicestatus.verify_conf_files: config json missing statusSource field in file: " + key)
                service_config_source = False

            try:
                service_config_type = self.json_configs[key]["config"]["statusType"]
            except KeyError:
                print("watchapservicestatus.verify_conf_files: config json missing serviceStatusInterval field"
                      " in file: " + key)
                service_config_type = False

            try:
                service_config_name = self.json_configs[key]["config"]["serviceName"]
            except KeyError:
                print("watchapservicestatus.verify_conf_files: config json missing serviceName field in file: " + key)
                service_config_name = False

            if not service_config_source:
                print("watchapservicestatus.verify_conf_files: config json has invalid type or value for statusSource field"
                      " in file: " + key)
                return False

            if service_config_source == "systemd":
                print("watchapservicestatus.verify_conf_files: config json states this is a statusSource=systemd type watchdog"
                      " in file: " + key)
                pass
            elif service_config_source == "docker":
                print("watchapservicestatus.verify_conf_files: config json states this is a statusSource=docker type watchdog"
                      " in file: " + key)
                pass

            else:
                print("watchapservicestatus.verify_conf_files: config json has invalid type or value for statusSource field"
                      " in file: " + key)
                return False

            if not service_config_type or not service_config_name:
                print("watchapservicestatus.verify_conf_files: config json missing statusType or serviceName field"
                      " in file: " + key)
                return False

        return True

    # Query the service status based on service type and get status values
    def get_service_status(self) -> bool:

        for key, value in self.json_configs.items():
            service_config_source = self.json_configs[key]["config"]["statusSource"]
            service_config_type = self.json_configs[key]["config"]["statusType"]
            service_config_name = self.json_configs[key]["config"]["serviceName"]

            # Store results of querying service status
            result_values = ""
            service_output_result = ""
            service_output_loadstate = ""
            service_output_activestate = ""
            service_output_substate = ""

            # If a service is broken report it as having error status
            # If this is a unit test return False to indicate failure
            error = False
            service_description = ""

            # Check systemd type services
            if service_config_source == "systemd":
                if self.test_status:
                    if service_config_type == "daemon":
                        result_values = ['Result=success', 'LoadState=loaded', 'ActiveState=active', 'SubState=running', '']
                    elif service_config_type == "oneshot":
                        result_values = ['Result=success', 'LoadState=loaded', 'ActiveState=active', 'SubState=exited', '']
                    elif service_config_type == "script":
                        result_values = ['Result=success', 'LoadState=loaded', 'ActiveState=activating', 'SubState=auto-restart', '']

                else:
                    command = "systemctl show " + service_config_name + ".service | egrep 'LoadState|ActiveState|SubState|^Result='"

                    try:
                        result = subprocess.check_output(command, timeout=10, universal_newlines=True, shell=True)
                    except subprocess.TimeoutExpired:
                        result = ""
                    except subprocess.CalledProcessError:
                        result = ""

                    result_values = result.split("\n")

                # print(result_values)

                for v in result_values:
                    if v:
                        if v.split("=")[0] == "Result":
                            service_output_result = v.split("=")[1]
                        elif v.split("=")[0] == "LoadState":
                            service_output_loadstate = v.split("=")[1]
                        elif v.split("=")[0] == "ActiveState":
                            service_output_activestate = v.split("=")[1]
                        elif v.split("=")[0] == "SubState":
                            service_output_substate = v.split("=")[1]

                if not service_output_result == "success":
                    service_description = service_config_name + " has unexpected Result (Result=" + service_output_result + ")"
                    error = True

                if not service_output_loadstate == "loaded":
                    service_description = service_config_name + " has unexpected LoadState (LoadState=" + service_output_loadstate + ")"
                    error = True

                if service_config_type == "daemon":
                    if not service_output_activestate == "active":
                        service_description = service_config_name + " has unexpected ActiveState (ActiveState=" + service_output_activestate + ")"
                        error = True

                    if not service_output_substate == "running":
                        service_description = service_config_name + " has unexpected SubState (SubState=" + service_output_substate + ")"
                        error = True

                if service_config_type == "script":
                    if (not service_output_activestate == "active") and (not service_output_activestate == "activating"):
                        service_description = service_config_name + " has unexpected ActiveState (ActiveState=" + service_output_activestate + ")"
                        error = True

                    if (not service_output_substate == "running") and (not service_output_substate == "auto-restart"):
                        service_description = service_config_name + " has unexpected SubState (SubState=" + service_output_substate + ")"
                        error = True

                if error:
                    self.reported_status[service_config_name] = {"serviceStatus": "Warning",
                                                                 "serviceDescription": service_description}

                    if self.test_status:
                        return False

                else:
                    self.reported_status[service_config_name] = {"serviceStatus": "Ok",
                                                                 "serviceDescription": "Ok"}
            # Check docker container status
            elif service_config_source == "docker":
                if self.test_status:
                    result_string = "true"
                else:
                    command = "docker inspect -f {{.State.Running}} " + service_config_name

                    try:
                        result = subprocess.check_output(command, timeout=10, universal_newlines=True, shell=True)
                    except subprocess.TimeoutExpired:
                        result = ""
                    except subprocess.CalledProcessError:
                        result = ""

                    result_string = result.strip()

                if result_string == "true":
                    service_description = "Ok"
                    self.reported_status[service_config_name] = {"serviceStatus": "Ok",
                                                                 "serviceDescription": service_description}
                else:
                    service_description = "Down"
                    self.reported_status[service_config_name] = {"serviceStatus": "Warning",
                                                                 "serviceDescription": service_description}
                    if self.test_status:
                        return False
            else:
                print("watchapservicestatus.get_service_status: status source does not match valid source")
                return False

        return True

    # Print out the reported service values
    def print_reported_status(self) -> bool:
        try:
            print("\nService Reported Statuses:")
            print(json.dumps(self.reported_status, indent=4, sort_keys=True))
        except Exception as e:
            print("watchapservicestatus.print_dm_report: Exception: " + str(e))
            return False

        return True

    # Create DM report out of the information returned
    def build_dm_report(self) -> bool:
        self.dm_output = io.StringIO()

        for service, value in self.reported_status.items():
            line = [self.hostname, self.reported_status[service]["serviceStatus"], service,
                    self.reported_status[service]["serviceDescription"]]

            try:
                self.dm_output.write(",".join(line) + '\r\n')
            except Exception as e:
                print("watchapservicestatus.build_dm_report: Exception: " + str(e))
                return False

        return True

    # Make sure the DM report is properly formatted
    def verify_dm_report(self) -> bool:
        report = self.dm_output.getvalue().split("\r\n")

        for line in list(filter(None, report)):
            results = line.split(",")

            # Each line must have 4 columns separated by commas
            # Format: CO4PHY050080545,Ok,Linux1,Linux is working fine
            if not len(results) == 4:
                print("watchapservicestatus.verify_dm_report: Number of expected columns is incorrect")
                return False

        return True

    # Write DM report to a tmp file
    def write_dm_report_to_disk(self) -> bool:
        if not os.path.isdir(self.watchdog_csv_dir):
            print("watchapservicestatus.write_dm_report_to_disk: Directory: " + self.watchdog_csv_dir + " not found")
        else:
            try:
                csv_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_tmp_file)

                with open(csv_file_path, "w") as text_file:
                    text_file.write(self.dm_output.getvalue())

            except Exception as e:
                print("watchapservicestatus.write_dm_report_to_disk: Exception: " + str(e))
                return False

        return True

    # Verify that the tmp file on disk is the same as the data that was supposed to be written
    def verify_write_dm_report_to_disk(self) -> bool:
        csv_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_tmp_file)

        if os.path.exists(csv_file_path):
            with open(csv_file_path, 'r') as content_file:
                content = content_file.read()

                content = content.split("\n")
                dm_output = self.dm_output.getvalue().split("\r\n")

                if list(filter(None, content)) == list(filter(None, dm_output)):
                    return True
                else:
                    print("watchapservicestatus.verify_write_dm_report_to_disk: csv content does not match dm_output content")
                    return False

        else:
            return False

    # Atomically rename the .tmp file to a .csv file which will be ingested by machinelocalwatchdog.py
    def atomically_rename_dm_report_on_disk(self) -> bool:
        tmp_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_tmp_file)
        csv_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_csv_file)

        # print("\nRenaming: " + tmp_file_path + " to: " + csv_file_path)

        if platform.system() == "Windows" and self.test_status:
            return True

        try:
            os.rename(tmp_file_path, csv_file_path)

        except Exception as e:
            print("watchapservicestatus.atomically_rename_dm_report_on_disk: Exception: " + str(e))
            return False

        return True

    # Print out the DM report
    def print_dm_report(self) -> bool:

        output = "\n".join(self.dm_output.getvalue().split("\r\n"))
        self.dm_output.close()

        try:
            print("\nDM Report:")
            print(output)
        except Exception as e:
            print("watchapservicestatus.print_dm_report: Exception: " + str(e))
            return False

        return True

    # Send out statsd data to MDM
    def send_statsd_report(self) -> bool:
        for service, value in self.reported_status.items():
            # Report status of 1 for OK
            if self.reported_status[service]["serviceStatus"] == "Ok":
                service_status = 1
            # Report Error to MDM for Warnings so we can investigate
            elif self.reported_status[service]["serviceStatus"] == "Warning":
                service_status = 2
            # Report Error to MDM for legitimate service failures
            elif self.reported_status[service]["serviceStatus"] == "Error":
                service_status = 2
            # Any status that is not on this list is not good so this acts as a catchall
            # for other possible failure scenarios, report Error to MDM
            else:
                service_status = 2

            # Send multi dim metric
            if not self.test_status:
                if not self.metrics.send_gauge(self.namespace, service + "Status", service_status):
                    return False

        return True

    # Execute all the methods in order
    def main(self):
        # Variable keeps track of where the watchdog is failing and is used for reporting
        watchdog_error = ""

        if not self.read_conf_files():
            print("watchapservicestatus.main: Could not read conf files for services (main.read_conf_files)")
            if not watchdog_error:
                watchdog_error = "Could not read conf files for services (main.read_conf_files)"

            # Create Warning report in DM due to failure
            self.reported_status["watchapservice"] = {"serviceStatus": "Warning",
                                                      "serviceDescription": "Could not read conf files for services (main.read_conf_files)"}

        if not self.verify_conf_files():
            print("watchapservicestatus.main: Could not verify conf files for services (main.verify_conf_files)")
            if not watchdog_error:
                watchdog_error = "Could not verify conf files for services (main.verify_conf_files)"

            # Create Warning report in DM due to failure
            self.reported_status["watchapservice"] = {"serviceStatus": "Warning",
                                                      "serviceDescription": "Could not verify conf files for services (main.verify_conf_files)"}

        if not self.get_service_status():
            print("watchapservicestatus.main: Could not read status files for some (or all) services (main.get_service_status)")
            if not watchdog_error:
                watchdog_error = "Could not read status files for some (or all) services (main.get_service_status)"

            # Create Warning report in DM due to failure
            self.reported_status["watchapservice"] = {"serviceStatus": "Warning",
                                                      "serviceDescription": "Could not read status files for some (or all) services (main.get_service_status)"}

        if not self.build_dm_report():
            print("watchapservicestatus.main: Could not build DM report (main.build_dm_report)")
            if not watchdog_error:
                watchdog_error = "Could not build DM report (main.build_dm_report)"

        if not self.verify_dm_report():
            print("watchapservicestatus.main: Could not verify DM report (main.verify_dm_report)")
            if not watchdog_error:
                watchdog_error = "Could not verify DM report (main.verify_dm_report)"

        if not self.write_dm_report_to_disk():
            print("watchapservicestatus.main: Could not write DM report to disk (main.write_dm_report_to_disk)")
            if not watchdog_error:
                watchdog_error = "Could not write DM report to disk (main.write_dm_report_to_disk)"

        if not self.verify_write_dm_report_to_disk():
            print("watchapservicestatus.main: Could not verify DM report written to disk (main.verify_write_dm_report_to_disk)")
            if not watchdog_error:
                watchdog_error = "Could not verify DM report written to disk (main.verify_write_dm_report_to_disk)"

        if not self.atomically_rename_dm_report_on_disk():
            print("watchaphdd.main: Could not atomically rename dm report file on disk (main.atomically_rename_dm_report_on_disk)")
            if not watchdog_error:
                watchdog_error = "Could not atomically rename dm report file on disk (main.atomically_rename_dm_report_on_disk)"

        # if not self.print_dm_report():
        #     print("watchapservicestatus.main: Could not print DM report (main.print_dm_report)")
        # if not watchdog_error:
        #     watchdog_error = "Could not print DM report (main.print_dm_report)"

        # Prepare status for MDM upload
        if watchdog_error:
            self.reported_status["watchapservice"] = {"serviceStatus": "Error",
                                                      "serviceDescription": watchdog_error}

        else:
            self.reported_status["watchapservice"] = {"serviceStatus": "Ok",
                                                      "serviceDescription": "Ok"}

        if not self.send_statsd_report():
            print("watchapservicestatus.main: Could not send statsd report (main.send_statsd_report)")
            sys.exit(1)

        sys.exit(0)  # EX_OK = everything executed normally

if __name__ == '__main__':

    configpath = "/ap/data/watchapservicestatus/conf.d/"
    watchdog_csv_dir = "/ap/data/watchdogoutput/csv_sources"
    watchdog_tmp_file = "watchapservicestatus.tmp"
    watchdog_csv_file = "watchapservicestatus.csv"

    WatchAPServiceStatus(configpath, watchdog_csv_dir, watchdog_tmp_file, watchdog_csv_file).main()
