#!/usr/bin/env python3

# Watchdog used to check version information and report on it

import sys
import os
import socket
import io
import datetime
import subprocess

from collections import OrderedDict

# if script's parent directory isn't in sys.path, then add it
if ('/'.join(sys.path[0].split("/")[:-1]) not in sys.path):
    sys.path.append('/'.join(sys.path[0].split("/")[:-1]))

from apcommon import apthelper


class WatchAPVersion(object):
    def __init__(self, watchdog_csv_dir: str, watchdog_tmp_file: str, watchdog_csv_file: str,
                 osversion_file: str="", update_success_file: str="", upgrade_success_file: str="", reboot_required_file: str=""):

        # Path to OS Files
        self.osversion_file = osversion_file
        self.update_success_file = update_success_file
        self.upgrade_success_file = upgrade_success_file
        self.reboot_required_file = reboot_required_file

        # machinelocalwatchdog folder and files to use for DM report
        self.watchdog_csv_dir = watchdog_csv_dir
        self.watchdog_tmp_file = watchdog_tmp_file
        self.watchdog_csv_file = watchdog_csv_file

        # Get hostname
        self.hostname = socket.gethostname()

        # List of debian packages to get version of
        self.debian_packages = ["aplinux-apwaltz", "aplinux-service-manager", "aplinux-shared-services",
                                "aplinux-geneva-mdsd-azsecpack", "aplinux-geneva-mdsd-syslog", "aplinux-geneva-mdmstack"]

        # Prepare dict for version result data
        self.version_results = OrderedDict()

    # Get the OS version by querying the file placed in the image in
    # /ap/data/OsVersion  This file not existing is an error since it should be imaged
    def get_os_version(self) -> bool:
        if os.path.isfile(self.osversion_file):
            try:
                with open(self.osversion_file, 'r') as f:
                    first_line = f.readline()
                    f.close()

            except Exception as e:
                print("\nwatchapversion: Exception: " + str(e))
                return False

            # The only supported OS right now is Ubuntu so we match that to make sure it's valid
            if "Ubuntu" in first_line:
                self.version_results["OSImageVersion"] = first_line.strip()
            else:
                print("watchapversion.get_os_version OsVersion does not match Ubuntu")
                return False
        else:
            print("watchapversion.get_os_version OsVersion file does not exist")
            return False

        return True

    # Get information on when the OS was last updated and if it requires reboot
    def get_os_update_info(self) -> bool:
        update_info = ""

        if os.path.isfile(self.update_success_file):
            try:
                update_info += "Last apt-get update: " + str(datetime.datetime.fromtimestamp(os.path.getmtime(self.update_success_file)).strftime("%Y-%m-%d %H:%M:%S"))
            except OSError:
                update_info += "Last apt-get update: Unknown"

        update_info += " | "

        if os.path.isfile(self.upgrade_success_file):
            try:
                update_info += "Last apt-get upgrade: " + str(datetime.datetime.fromtimestamp(os.path.getmtime(self.upgrade_success_file)).strftime("%Y-%m-%d %H:%M:%S"))
            except OSError:
                update_info += "Last apt-get upgrade: Unknown"

        update_info += " | "

        if os.path.isfile(self.reboot_required_file):
            update_info += "Upgrade Reboot Required: Yes"
        else:
            update_info += "Upgrade Reboot Required: No"

        self.version_results["OSUpdateInformation"] = update_info

        return True

    # Use apthelper.py library to make sure package is installed before checking version
    def check_package_installed(self, package) -> bool:
        if apthelper.apt_get_is_installed(package):
            return True
        else:
            return False

    # Use apthelper.py library to query package versions
    def get_package_version(self, package) -> str:
        if self.check_package_installed(package):
            version = apthelper.apt_get_version(package)
            return version.strip()
        else:
            return ""

    # Query package versions and store them in results dictionary
    def get_shared_services_version(self) -> bool:
        for package in self.debian_packages:
            version = self.get_package_version(package)

            self.version_results[package] = version

        return True

    # Get existing kernel version from the system
    def get_kernel_version(self) -> str:
        command = "uname -a"

        try:
            result_kernel = subprocess.check_output(command, timeout=10, universal_newlines=True, shell=True)
        except subprocess.TimeoutExpired:
            return ""
        except subprocess.CalledProcessError:
            return ""

        return result_kernel.strip()

    # Process kernel data and store in results dictionary
    def get_kernel_data(self) -> bool:
        self.version_results["KernelVersion"] = self.get_kernel_version()

        return True

    # Print out version results
    def print_version_results(self):
        try:
            print("\nVersions of OS, Kernel and Linux Shared Services:")
            print(self.version_results)

        except Exception as e:
            print("watchapversion.print_versions: Exception: " + str(e))
            return False

        return True

    # Create DM report out of the information returned
    def build_dm_report(self):
        self.dm_output = io.StringIO()

        for key, value in self.version_results.items():
            status = "Ok"

            if not self.version_results[key]:
                status = "Warning"

            line = [self.hostname, status, key, value]

            try:
                self.dm_output.write(",".join(line) + '\r\n')
            except Exception as e:
                print("watchapversion.build_dm_report: Exception: " + str(e))
                return False

        return True

    # Make sure the DM report is properly formatted
    def verify_dm_report(self):
        report = self.dm_output.getvalue().split("\r\n")

        for line in list(filter(None, report)):
            results = line.split(",")

            # Each line must have 4 columns separated by commas
            # Format: CCPSCH010261732,ok,smartTst:2:0,HDD disk 2:0 model 'MB2000GDMTE' serial 'Z1X0G381' firmware 'HPGC' is OK
            if not len(results) == 4:
                print("watchapversion.verify_dm_report: Number of expected columns is incorrect")
                return False

        return True

    # Write DM report to a tmp file
    def write_dm_report_to_disk(self) -> bool:
        if not os.path.isdir(self.watchdog_csv_dir):
            print("watchapversion.write_dm_report_to_disk: Directory: " + self.watchdog_csv_dir + " not found")
            return False
        else:
            csv_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_tmp_file)

            try:
                with open(csv_file_path, "w") as text_file:
                    text_file.write(self.dm_output.getvalue())
            except Exception as e:
                print("watchapversion.write_dm_report_to_disk: Exception: " + str(e))
                return False

        return True

    # Verify that the tmp file on disk is the same as the data that was supposed to be written
    def verify_write_dm_report_to_disk(self) -> bool:
        csv_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_tmp_file)

        if os.path.exists(csv_file_path):
            try:
                with open(csv_file_path, 'r') as content_file:
                    content = content_file.read()
            except Exception as e:
                print("watchapversion.verify_write_dm_report_to_disk: Exception: " + str(e))
                return False

            content = content.split("\n")
            dm_output = self.dm_output.getvalue().split("\r\n")

            if list(filter(None, content)) == list(filter(None, dm_output)):
                return True
            else:
                print("watchapversion.verify_write_dm_report_to_disk: csv content does not match dm_output content")
                return False

        else:
            return False

    # Atomically rename the .tmp file to a .csv file which will be ingested by machinelocalwatchdog.py
    def atomically_rename_dm_report_on_disk(self) -> bool:
        tmp_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_tmp_file)
        csv_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_csv_file)

        try:
            os.rename(tmp_file_path, csv_file_path)

        except Exception as e:
            print("watchapversion.atomically_rename_dm_report_on_disk: Exception: " + str(e))
            return False

        return True

    # Print out the DM report
    def print_dm_report(self):

        output = "\n".join(self.dm_output.getvalue().split("\r\n"))
        self.dm_output.close()

        try:
            print("\nDM Report:")
            print(output)
        except Exception as e:
            print("watchapversion.print_dm_report: Exception: " + str(e))
            return False

        return True

    # Execute all the methods in order
    def main(self):
        if not self.get_os_version():
            print("watchapversion.main: Could not get OS Version (main.get_os_version)")
            sys.exit(1)

        if not self.get_os_update_info():
            print("watchapversion.main: Could not get Update Information (main.get_os_update_info)")
            sys.exit(1)

        if not self.get_shared_services_version():
            print("watchapversion.main: Could not get shared services version (main.get_shared_services_version)")
            sys.exit(1)

        if not self.get_kernel_data():
            print("watchapversion.main: Could not get kernel data (main.get_kernel_data)")
            sys.exit(1)

        # if not self.print_version_results():
        #     print("watchapversion.main: Could not print version results (main.print_version_results)")
        #     sys.exit(1)

        if not self.build_dm_report():
            print("watchapversion.main: Could not build DM report (main.build_dm_report)")
            sys.exit(1)

        if not self.verify_dm_report():
            print("watchapversion.main: Could not verify DM report (main.verify_dm_report)")
            sys.exit(1)

        if not self.write_dm_report_to_disk():
            print("watchapversion.main: Could not write DM report to disk (main.write_dm_report_to_disk)")
            sys.exit(1)

        if not self.verify_write_dm_report_to_disk():
            print("watchapversion.main: Could not verify DM report written to disk (main.verify_write_dm_report_to_disk)")
            sys.exit(1)

        if not self.atomically_rename_dm_report_on_disk():
            print("watchapversion.main: Could not atomically rename dm report file on disk (main.atomically_rename_dm_report_on_disk)")
            sys.exit(1)

        # if not self.print_dm_report():
        #     print("watchapversion.main: Could not print DM report (main.print_dm_report)")
        #     sys.exit(1)

        sys.exit(0)  # EX_OK = everything executed normally

if __name__ == '__main__':
    watchdog_csv_dir = "/ap/data/watchdogoutput/csv_sources"
    watchdog_tmp_file = "watchapversion.tmp"
    watchdog_csv_file = "watchapversion.csv"
    osversion_file = "/ap/data/OsVersion"
    update_success_file = "/var/lib/apt/periodic/update-success-stamp"
    upgrade_success_file = "/var/lib/apt/periodic/upgrade-stamp"
    reboot_required_file = "/var/run/reboot-required"

    WatchAPVersion(watchdog_csv_dir, watchdog_tmp_file, watchdog_csv_file, osversion_file, update_success_file,
                   upgrade_success_file, reboot_required_file).main()
